# Copyright (C) 2002-2014 NEC Corporation
# All Rights Reserved.
#
#  [See ftdisk]

#
$ENV{LC_ALL}= 'POSIX';
$FTDISK_PROC_SCSI_SCSI= '/proc/scsi/scsi';
$FTDISK_ROOTDIR= '/opt/nec/ftras';
$FTDISK_BINDIR= "$FTDISK_ROOTDIR/sbin";

#
(
 $SCSI_DB_SERIAL2PATH_H,
 $SCSI_DB_PATH2SERIAL_H,
 $SCSI_DB_DEVNAME2SERIAL_H,
 $SCSI_DB_DEVNAME2PATH_H,
 $SCSI_DB_SERIAL2INFO_H,
) = (0..4);

#
sub kdev2udev{
    my($kd)= @_;

    if(!open(UDEVINFO, "/sbin/udevadm info --root --query=name --path=/sys/block/$kd |")){
        return undef;
    }
    while(<UDEVINFO>){
        if(/(sd[a-z]+)/){
            close(UDEVINFO);
            return $1;
        }
    }
    
    close(UDEVINFO);
    return $kd;
}

sub udev2kdev{
    my($ud)= @_;
    
    if(!open(UDEVINFO, "/sbin/udevadm info --query=path --name=/dev/$ud |")){
        return undef;
    }
    while(<UDEVINFO>){
        if(/(sd[a-z]+)/){
            close(UDEVINFO);
            return $1;
        }
    }
    
    close(UDEVINFO);
    return $ud;
}

#
sub scsiSlot2Path{
    my($slt1)= @_;
    my(@paths);

    my(@db)= &scsiGetSimpleDB($SCSI_DB_DEVNAME2PATH_H);
    $paths[$#paths+1]= $db[0]->{&scsiSlot2Devname($slt1)};

    return @paths;
}

#
sub scsiExtractLivePaths{
    my(@paths)= @_;

    my(@lives);
    my(%pth2ser)= %{(&scsiGetSimpleDB($SCSI_DB_PATH2SERIAL_H))[0]};
    foreach (@paths){
        $lives[$#lives+1]= $_ if($pth2ser{$_});
    }

    return @lives;
}

#
sub scsicmp{
    my(@a)= split(/\D+/, $a);
    my(@b)= split(/\D+/, $b);

    foreach (0..$#a){
        my($c)= $a[$_] <=> $b[$_];
        if($c){ return $c; }
    }
}

#
sub scsiDevname2Slot{
    my($dn)= @_;

    if($dn =~ /sd([a-p])/){
        return index(" abcdefghijklmnop", $1);
    }
    else{
        return undef;
    }
}

#
sub scsiSlot2Devname{
    my($slt)= @_;

    if(($slt >= 1) and ($slt <= 16)){
        return ('sda', 'sdb', 'sdc', 'sdd', 'sde', 'sdf', 'sdg', 'sdh',
                'sdi', 'sdj', 'sdk', 'sdl', 'sdm', 'sdn', 'sdo', 'sdp')[$slt-1];
    }
    else{
        return undef;
    }
}

#
sub scsiDevname2HWpath{
    my($dn)= @_;
    my(%hw_path)= (
        'sda' => '10/40/1',
        'sdb' => '10/40/2',
        'sdc' => '10/40/3',
        'sdd' => '10/40/4',
        'sde' => '10/40/5',
        'sdf' => '10/40/6',
        'sdg' => '10/40/7',
        'sdh' => '10/40/8',
        'sdi' => '11/40/1',
        'sdj' => '11/40/2',
        'sdk' => '11/40/3',
        'sdl' => '11/40/4',
        'sdm' => '11/40/5',
        'sdn' => '11/40/6',
        'sdo' => '11/40/7',
        'sdp' => '11/40/8',
    );
     
    return $hw_path{$dn} if($dn =~ /^sd[a-p]$/);
}

#
sub scsiDevname2Serial{
    my($dn)= @_;

    my($kdn)= &udev2kdev($dn);
    my(@scsi_info)= split(/\s+/, `/sbin/scsi_id --whitelisted --page=0x80 --device=/dev/$dn`);
    
    return $scsi_info[2];
}

#
sub scsiDevname2Vender{
    my($dn)= @_;
    
    my($kdn)= &udev2kdev($dn);
    if($kdn){
        if(open(VENDOR, "/sys/block/$kdn/device/vendor")){
            my($vndr)= <VENDOR>;
            $vndr=~ s/\s*\n//;
            close(VENDOR);
            return $vndr;
        }
    }
    return "-";
}

#
sub scsiDevname2Model{
    my($dn)= @_;

    my($kdn)= &udev2kdev($dn);
    if($kdn){
        if(open(MODEL, "/sys/block/$kdn/device/model")){
            my($mdl)= <MODEL>;
            $mdl=~ s/\s*\n//;
            close(MODEL);
            return $mdl;
        }
    }
    return "-";
}

#
sub scsiPath2Devname{
    my($pth)= @_;

    if($pth=~ /^h(\d+)c(\d+)t(\d+)l(\d+)$/){
        $pth= "${1}:${2}:${3}:${4}";
    }
    if($pth=~ /^(\d+):(\d+):(\d+):(\d+)$/){
        foreach(`/bin/ls -1 /sys/class/scsi_device/$pth/device/block 2> /dev/null`){
            if(/^(sd\w+)$/){
                return &kdev2udev($1);
            }
        }
    }
    return undef;
}

#
sub scsiGetSimpleDB{
    my(@kinds)= @_;

    local(%dn2ser,  %ser2info);
    local(%ser2pth, %pth2ser);
    local(%dn2pth);

    my($dn, $ser, $pth);
    if(open(LSSCSI, "ls -1 /sys/class/scsi_device |")){
        while(<LSSCSI>){
            chomp($pth= $_);
            if($pth=~ /^(\d+):(\d+):(\d+):(\d+)$/){
                $pth= &normalize("h${1}c${2}t${3}l${4}");
                $dn= &scsiPath2Devname($pth);
                next if(!$dn);
                $ser= &scsiDevname2Serial($dn);
                $dn2ser{$dn}= $ser;
                if($pth=~ /^h(\d+)c(\d+)t(\d+)l(\d+)$/){
                    $dn2pth{$dn}= $pth;
                    $pth2ser{$pth}= $ser;
                    $ser2pth{$ser}= $pth;
                }
                $ser2info{$ser}= [$dn,
                                  &isMounted($dn),
                                  &scsiDevname2Vender($dn),
                                  &scsiDevname2Model($dn)];
            }
            #elsif($line=~ /timeout=(\d+)/){
            #    push(@{$ser2info{$ser}}, $1);
            #}
        }
        close(LSSCSI);
    }

    my(@db);
    foreach (@kinds){
        if($_ eq $SCSI_DB_SERIAL2PATH_H)   { $db[$#db+1]= \%ser2pth;   }
        elsif($_ eq $SCSI_DB_PATH2SERIAL_H)   { $db[$#db+1]= \%pth2ser;   }
        elsif($_ eq $SCSI_DB_DEVNAME2SERIAL_H){ $db[$#db+1]= \%dn2ser;    }
        elsif($_ eq $SCSI_DB_DEVNAME2PATH_H)  { $db[$#db+1]= \%dn2pth;    }
        elsif($_ eq $SCSI_DB_SERIAL2INFO_H)   { $db[$#db+1]= \%ser2info;  }
    }

    return @db;
}

#
sub normalize{
    my($s)= @_;

    { redo if($s =~ s/(^|\D)0+(\d+)/$1$2/); }

    return $s;
}

##
sub getMountInfo{
    if(!open(MOUNT, "/bin/mount |")){
        return ();
    }
    my(%mntinfo);
    while(<MOUNT>){
        my(@a)= split;
        $mntinfo{$a[0]}= $a[2];
    }
    close(MOUNT);

    return %mntinfo;
}

#
sub getMdstat{
    if(!open(MDSTAT, "/proc/mdstat")){
        return undef;
    }
    my(%mdstat, $md, $udev);
    my(@cond_from, @cond_to);

    while(<MDSTAT>){
        if(/^(md\d+)\s+:\s+(\w+)\s*(.*)$/){
            $md= $1;
            $mdstat{$md}{'stat'}= $2;
            if($3 =~ /\w*raid(\d+)\s*(.*)$/){
                $mdstat{$md}{'level'}= $1;
                my(@a);
                foreach (split(/\s+/, $2)){
                    if(/^(sd[a-z]+)(\d+\[(\d+)\].*)/){
                        $udev= &kdev2udev($1);
                        my($num) = &getRaidDeviceNo($md, $udev);
                        $cond_from[$num] = $num;
                        $cond_to[$num]   = $3;
                        $a[$3]= "$udev$2";
                    }
                    elsif(/^md\d+\[(\d+)\].*/){
                        $a[$1]= "$_";
                    }
                }
                $mdstat{$md}{'devices'}= [@a];
                $mdstat{$md}{'num'}= $num;
                chomp($mdstat{$md}{'sync_action'}= `cat /sys/block/$md/md/sync_action 2>/dev/null`);
                chomp($mdstat{$md}{'sync_speed_min'}= `cat /sys/block/$md/md/sync_speed_min 2>/dev/null`);
                chomp($mdstat{$md}{'sync_speed_max'}= `cat /sys/block/$md/md/sync_speed_max 2>/dev/null`);
            }
        }
        elsif(/^\s+(\d+)\s+blocks.+?\[(\d+\/\d+)\]\s+\[([U_]+)\]$/){
            $mdstat{$md}{'blocks'}= $1;
            $mdstat{$md}{'cond'}= [split(//, $3)];

            my($i);
            foreach $i (0, 1){
                if($cond_from[$i] != $cond_to[$i]){
                    $mdstat{$md}{'cond'}[$cond_to[$i]] = $mdstat{$md}{'cond'}[$cond_from[$i]];
                    $cond_from[$i] = undef;
                    $cond_to[$i] = undef;
                }
            }
        }
        elsif(/^\s+(\d+)\s+blocks.+?(\d+)k\s+chunks$/){
            $mdstat{$md}{'blocks'}= $1;
            $mdstat{$md}{'chunk'}= $2;
        }
        elsif(/^\s+bitmap:\s+(\d+\/\d+\s+pages\s+\[\d+KB\])\,\s+(\d+)KB\s+chunk$/){
            $mdstat{$md}{'bitmap'}= $1;
            $mdstat{$md}{'chunk'}= $2;
        }
        elsif(/recovery\s*=\s*(\S+)/){
            $mdstat{$md}{'recovery'}= $1;
        }
        elsif(/resync\s*=\s*(\S+)/){
            $mdstat{$md}{'resync'}= $1;
        }
        elsif(/check\s*=\s*(\S+)/){
            $mdstat{$md}{'resync'}= $1;
        }
    }
    close(MDSTAT);

    return %mdstat;
}

#
sub getRaidDeviceNo{
    my($md, $udev) = @_;
    my($num);

    if(!open(MDADMDETAIL, "/sbin/mdadm --detail /dev/$md |")){
        return undef;
    }
    while(<MDADMDETAIL>){
        if(/^\s*\d*\s*\d*\s*\d*\s*(\d*)\s*.*\/dev\/(sd[a-z]*)\d*\s*$/){
            if($udev eq $2){
                $num = $1;
                last;
            }
        }
    }
    close(MDADMDETAIL);

    return $num;
}

#
sub setMdFastSync{
    my($md, $act)= @_;
    
    if($act eq "on"){
        return system("/sbin/mdadm --grow --bitmap=internal /dev/$md");
    }
    elsif($act eq "off"){
        return system("/sbin/mdadm --grow --bitmap=none /dev/$md");
    }
    else{
        return -1;
    }
}

#
sub setMdSyncSpeed{
    my($md, $act, $val)= @_;
    
    my(%mdstat)= &getMdstat();

    if($act eq "min" and $val < $mdstat{$md}{sync_speed_max}){
        return system("/bin/echo -n $val > /sys/block/$md/md/sync_speed_min");
    }
    elsif($act eq "max" and $val > $mdstat{$md}{sync_speed_min}){
        return system("/bin/echo -n $val > /sys/block/$md/md/sync_speed_max");
    }
    else{
        return -1; # failed to set a parameter.
    }
}

#
sub setMdSyncAction{
    my($md, $act)= @_;

    if($act =~ /^(idle|resync|recover|check|repair)$/){
        return system("/bin/echo $act > /sys/block/$md/md/sync_action");
    }
    else{
        return -1; # failed to set a parameter.
    }
}

#
sub getPosOfMdstat{
    my($dn, @array)= @_;

    if($dn !~ /\d+$/){
        $dn .= '\d+';
    }
    my($i)= scalar @array;
    while($i-- > 0){  #reverse search is efficient here
        if($array[$i] =~ /^$dn\[(\d+)\]/){
            return $1;
        }
    }

    return -1;
}

#
sub isRaided{
    my($dn)= @_;

    my(%mdstat)= &getMdstat();
    foreach $md (keys %mdstat){
        if(&getPosOfMdstat($dn, @{$mdstat{$md}{'devices'}}) >= 0){
            return 1;
        }
    }

    return 0;
}

#
sub isMounted{
    my($dn)= @_;
    my($num)= 0;

    my(%mntinfo)= &getMountInfo();
    my(%mdstat)= &getMdstat();
    foreach $mntdev (keys %mntinfo){
        if(($mntdev =~ /$dn\d*/) or
           (($mntdev =~ s/^\/dev\/(md\d+)$/$1/g) and
            &getPosOfMdstat($dn, @{$mdstat{$mntdev}{'devices'}}) >= 0)){
                $num++;
        }
    }
    return $num;
}

##
sub isMtbf{
    my($path)= @_;

    if(open(REASON, "/opt/ft/bin/ftsmaint ls $path | grep Reason |")){
        while(<REASON>){
            if(/^Reason\s*\:\s*(\w*)$/){
                if($1 =~ /(\w*MTBF\w*)/){
                    close(REASON);
                    return $1;
                }
            }
        }
        close(REASON);
    }
    return;
}

#
sub getPartitionTables{
    my($dn)= @_;
    if(!open(FDISK, "/sbin/fdisk -lu /dev/$dn |")){
        return ();
    }
    my(%ptables);
    while(<FDISK>){
        if(/^(\d+)\s*heads,\s*(\d+)\s*sectors\/track,\s*(\d+)\s*cylinders/){
            $ptables{'heads'}= $1;
            $ptables{'sectors'}= $2;
            $ptables{'cylinders'}= $3;
        }
        elsif(/^Units\s*=\s*sectors of\s*(\d+\s*\*\s*\d+)\s*=\s*(\d*)\s*bytes/){
            $ptables{'Units'}= $1;
        }
        elsif(/^(\/dev\/\S+?(\d+))\s+(\*?)\s+(\d+)\s+(\d+)\s+(\d+)\+?\s+(\w+)\s+(.+)/){
            $ptables{'device'}[$2]= [$1, $3, $4, $5, $6, $7, $8];
        }
    }
    close(FDISK);

    return %ptables;
}

#
sub copyPartitionTable{
    my($from, $to, $force)= @_;

    my($opt);  # --force is dangerous : DO NOT USE HERE
    my(%t)= &getPartitionTables($from);
    if($force == 1){
        $opt = "--force";
    }
    if($t){
        $opt .= " -H $t{'heads'}" if($t{'heads'});
        $opt .= " -S $t{'sectors'}" if($t{'sectors'});
    }

    return system("/sbin/sfdisk --force -R -d /dev/$from |
                   /sbin/sfdisk --force $opt /dev/$to");
#    return system("/sbin/sfdisk --force -R -d /dev/$from |
#                   /sbin/sfdisk $opt /dev/$to");
}

#
sub copyMBLoader{
    my($from, $to)= @_;

    my(%t)= &getPartitionTables($from);
    if($t{'sectors'}){
        my($count)= $t{'sectors'};
        foreach $n (1 .. 4){
            if($t{'device'}[$n]){
                if($t{'device'}[$n][2] < $count){
                    $count= $t{'device'}[$n][2];
                }
                last;
            }
        }
        if(--$count >= 1){
            system("/bin/dd if=/dev/$from skip=1 of=/dev/$to seek=1 bs=512 count=$count 2>/dev/null");
        }
    }

    return system("/bin/dd if=/dev/$from of=/dev/$to bs=446 count=1 2>/dev/null");
}

#
sub getNewMdname{
    my($seek_start)= @_;
    my(@mdlst);
    
    my(%mdstat)= &getMdstat();
    foreach $md (keys %mdstat){
        if($md =~ /^(md(\d+))/){
            $mdlst[$2]= $1;
        }
    }
    
    $seek_start= 0 unless($seek_start);
    my($num)= $seek_start;
    foreach($seek_start .. $#mdlst){
        return "md$_" if(!$mdlst[$_]);
        $num++
    }
    return "md$num";
}

#
sub isExist{
    my($mddev)= @_;
    
    my(%mdstat)= &getMdstat();
    foreach $md (keys %mdstat){
        if($md =~ /^$mddev/){
            return 1;
        }
    }
    return 0;
}

#
sub raidGetSlotBuddy{
    my($slt)= @_;

    if($slt >= 1){
        return $slt+8 if($slt <= 8);
        return $slt-8 if($slt <= 16);
    }

    return 0;
}

#
#
sub devName2No{
    my($dn)= @_;

    my($rdev)= (stat("/dev/$dn"))[6];

    return ($rdev >> 8, $rdev & 0xff);
}

#
sub devName2Partition{
    my($name)= @_;

    my($partition);
    do{
        last if(!open(MNT, "/etc/mtab"));  #/proc/mounts is NG.
        while(<MNT>){
            my(@a)= split;
            if($a[0] eq "/dev/$name"){
                $partition= $a[1];
                last;
            }
        }
        close(MNT);
        unless($partition){
            last if(!open(SWP, "/proc/swaps"));
            while(<SWP>){
                my(@a)= split;
                if($a[0] eq "/dev/$name"){
                    $partition= 'swap';
                    last;
                }
            }
            close(SWP);
        }
    }
    while(0);

    return $partition;
}

#
sub devPartition2Name{
    my($partition)= @_;

    my($name);
    do{
        last if(!open(MNT, "/etc/mtab"));
        while(<MNT>){
            my(@a)= split;
            if($a[1] eq $partition){
                $name= $a[0];
                $name =~ s/^\/dev\///;
                last;
            }
        }
        close(MNT)
    }
    while(0);

    return $name;
}

#
sub devLabel2Name{
    my($label, @patterns)= @_;

    if(!open(PART, '/proc/partitions')){
        return undef;
    }
    my($name);
    SEARCH: foreach $re (@patterns){
        foreach $dn (grep s/^.*?\s+($re\d)$/$1/, <PART>){
            my($l2)= &dev_name2label($dn);
            if($label eq $l2){
                chomp($name= $dn);
                last SEARCH;
            }
        }
        seek(PART, 0, 0);
    }
    close(PART);

    return $name;
}

#
sub devName2Label{
    my($name)= @_;

    chomp(my($label)= `e2label /dev/$name 2>/dev/null`);

    return $label;
}

#
sub getScsiDiskSize{
  my($slt)= @_;

  my($dn)= &scsiSlot2Devname($slt);
  if(!$dn){
      return -1;
  }
  my(%tables)= &getPartitionTables($dn);

  return int eval($tables{"Units"}) *
             $tables{"heads"} *
             $tables{"sectors"} *
             $tables{"cylinders"} / 1024 / 1024;
}

#
# @sizes= (0,S1,..,S5,..,-1)
#
sub makePartitionTable{
    my($dn, $type, @sizes)= @_;

    ##
    ## The difference of qla1280 between of installer and of ftlinux
    ## cause a kind of raid problem at re-creating a pair of raid1.
    ## Here, as a first aid, specifying '-H255' and '-S63' can avoid it.
    ##  (July 4, 2002)
    ##
    #if(!open(CMD, "| /sbin/sfdisk -uM -H255 -S63 /dev/$dn")){
    #    &error(1, "Cannot execute: /sbin/sfdisk");
    #}
    if(!open(CMD, "| /sbin/sfdisk -uM /dev/$dn")){
        &error(1, "Cannot execute: /sbin/sfdisk");
    }
    my($slice)= 1;
    if(@sizes > 5){
        for(; $slice < 5; ++$slice){
            if($sizes[$slice] > 0){
                print CMD ",$sizes[$slice],$type\n";
            }
            else{
                print CMD ",,5\n";  # as extended partition
                ++$slice;
                last;
            }
        }
        for(; $slice < 5; ++$slice){
            print CMD ";\n";
        }
    }
    for(; $slice < @sizes; ++$slice){
        if($sizes[$slice] <= 0){
            print CMD ",,$type\n" if($sizes[$slice] < 0);
            last;
        }
        print CMD ",$sizes[$slice],$type\n";
    }
    close(CMD);

    return ($?) ? -1 : 0;
}

#
sub makePartitionTable64{
    my($dn, $type, @sizes)= @_;

    ##
    ## The difference of qla1280 between of installer and of ftlinux
    ## cause a kind of raid problem at re-creating a pair of raid1.
    ## Here, as a first aid, specifying '-H255' and '-S63' can avoid it.
    ##  (July 4, 2002)
    ##
    #if(!open(CMD, "| /sbin/sfdisk -uM -H255 -S63 /dev/$dn")){
    #    &error(1, "Cannot execute: /sbin/sfdisk");
    #}
    if(!open(CMD, "| /sbin/sfdisk --force -uS /dev/$dn")){
        &error(1, "Cannot execute: /sbin/sfdisk");
    }
    my($slice)= 1;
    my($start) = 64;
    my($sizes_sectr);
    if(@sizes > 5){
        for(; $slice < 5; ++$slice){
            if($sizes[$slice] > 0){
                $sizes_sectr = $sizes[$slice] * 2048;
                print CMD "$start,$sizes_sectr,$type\n";
                $start = $start + $sizes_sectr;
            }
            else{
                print CMD "$start,,5\n";  # as extended partition
                ++$slice;
                last;
            }
        }
        for(; $slice < 5; ++$slice){
            print CMD ";\n";
        }
    }
    for(; $slice < @sizes; ++$slice){
        if($slice > 4){
            $start = $start + 64;
        }
        if($sizes[$slice] <= 0){
            print CMD "$start,,$type\n" if($sizes[$slice] < 0);
            last;
        }
        $sizes_sectr = $sizes[$slice] * 2048;
        print CMD "$start,$sizes_sectr,$type\n";
        $start = $start + $sizes_sectr;
    }
    close(CMD);

    return ($?) ? -1 : 0;
}

1;
